/***************************************************************************
           arrays-main.cpp  -  Example of array variables in DSLib v2
                             -------------------
    begin                : Tue Jun 13 2000
    copyright            : (C) 2003 by Ivan Martinez
    email                : imr@oersted.dtu.dk
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2.1 of the License, or   *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <dynamicsystem.h>
#include <blockvariable.h>
#include <function.h>
#include <periodicscheduler.h>
#include <loggingfifo.h>
#include <message.h>
#include <messages.h>
//#include <settingsharedmemory.h>
#include <fstream>
#include <comediinput.h>
#include <comedioutput.h>
#include <comedi.h>
#include <iostream>

// One second is 1e9 nanoseconds (time unit required by functions)
#define SECONDS 1e9
// Variables are logged in one every 1/LOGGING_SAMPLING_RATE executions
// of the scheduler
#define LOGGING_SAMPLING_RATE 0.5
// Realtime fifos will be able to store FIFOS_SAMPLE samples of their
// variable
#define FIFOS_LENGTH 100
// Period of the execution of the scheduler, in seconds
#define EXECUTION_PERIOD	0.01

// Constants for the A/D boards
// A/D board attached to /dev/comedi0
#define AD_COMEDI0	"/dev/comedi0"
// A/D board attached to /dev/comedi1 (see Makefile)
#define AD_COMEDI1	"/dev/comedi1"
// Subdevices
#define AD_ANALOG_INPUT	0
#define AD_ANALOG_OUTPUT	1
#define AD_COUNTER_TIMER	2
// Ranges
#define AD_RANGE_10_10	0	// -10V +10V
#define AD_RANGE_5_5	1	// -5V +5V
#define AD_RANGE_25_25 2	// -2.5V +2.5V
#define AD_RANGE_125_125	3	// -1.25V +1.25V
#define AD_RANGE_0_10	4	// 0V +10V
#define	AD_RANGE_0_5	5	// 0V +5V
#define AD_RANGE_0_25	6	// 0V +2.5V
#define AD_RANGE_0_125	7	// 0V +1.25V
// Reference constants defined in comedi.h mean the following to us:
// AREF_GROUND	Single ended input in our DAS boards
// AREF_COMMON	Same as AREF_GROUND
// AREF_DIFF	Differetial inputs in our DAS boards
// AREF_OTHER	Not supported by our these A/D boards	
#define AD_SE_CHANNELS	16	// Number of channels in single-ended mode
#define AD_DIFF_CHANNELS	8	// Number of channels in differential mode

// Constants for the D/A board
// DDA board in port 00:0C is attached to /dev/comedi3 (see Makefile)
#define DA_COMEDI3	"/dev/comedi3"
// Subdevices
#define DA_ANALOG_OUTPUT	0
#define DA_DIGITAL_IO1	1
#define DA_DIGITAL_IO2	2
// Ranges
#define DA_RANGE_10_10	0	// -10V +10V
#define DA_RANGE_5_5	1	// -5V +5V
#define DA_RANGE_25_25 2	// -2.5V +2.5V
#define DA_RANGE_0_10	3	// 0V +10V
#define	DA_RANGE_0_5	4	// 0V +5V
#define DA_RANGE_0_25	5	// 0V +2.5V
#define DA_CHANNELS	8	// Number of channels

//Devices to use
#define AD1	// DA board attached to /dev/comedi0
#define AD2	// DA board attached to /dev/comedi1
#define DA	// DA board attached to /dev/comedi3

#define VARIABLES_TYPE	"float"	// Type for the variables of the system

int main(int argc, char *argv[])
{
	// General parameters for the dynamic system
	CDynamicSystem DynamicSystem;
	// Maximum execution time
	DynamicSystem.SetTime((long long) (1000 * SECONDS));
	// Resolution for the realtime timer. This affects the precission of
	// the period and the times returned by reporting fifos. Small resolution
	// times increase CPU workload.
	DynamicSystem.SetClockResolution((long long) (0.0001 * SECONDS));
	
	// Parameters for the periodic scheduler. Every scheduler defines a
	// realtime thread that executes blocks in the same order they are
	// added to it.
	CPeriodicScheduler* pPSScheduler = new CPeriodicScheduler();
	// Schedulers are added to the dynamic system
	DynamicSystem.AddComponent(pPSScheduler);
	pPSScheduler->SetIdentifier("Scheduler");
	// Period of execution
	pPSScheduler->SetPeriod((long long) (EXECUTION_PERIOD * SECONDS));
	// Priority of the thread, to solve competition between realtime threads
	pPSScheduler->SetPriority(0);
	// We map the thread execution to CPU1
	/* pPSScheduler->SetCPU(1);	NOT IMPLEMENTED YET */

	/* NOT IMPLEMENTED YET
	// Reporting Fifo for pPSScheduler. Reporting Fifos report the start and
	// finish time for every execution of the realtime thread of its Loop.
	// Values are in RTIME format for RTAI and hrtime_t for RTLinux (see their
	// documentation). The start execution time is preceeded by a 0 and the
	// finish time by a 1 (both in RTAI/hrtime_t format too)
	CReportingFifo* pRFScheduler = new CReportingFifo();
	// The file will be "/dev/rtf0"
	pRFScheduler->SetNumber(0);
	// Fifos work as a buffer between the realtime and the user-space process.
	// We set the fifo to be able to store FIFOS_LENGTH samples.
	pRFScheduler->SetLength(FIFOS_LENGTH);
	// Logging fifos are added to variables. Block's inputs and outputs are
	// variables
	pPSScheduler->AddComponent(pRFScheduler);
	*/

	// Superblock that reads the inputs of AD_COMEDI0 and writes the values
	// to its output as an array of type VARIABLES_TYPE and length
	// AD_SE_CHANNELS.
	CSuperBlock* pSBAD1 = new CSuperBlock();
	// Blocks are added to schedulers or other superblocks. In this case,
	// pSScheduler.
	pPSScheduler->AddComponent(pSBAD1);
	// Identifier.
	pSBAD1->SetIdentifier("ad1");
	// Output. The third argument, when used, specifies that it's an array
	// and sets its length.
	pSBAD1->AddOutput("inputs_array", VARIABLES_TYPE, AD_SE_CHANNELS);

	// Comedi Input Block to read channel 0 of AD_COMEDI0
	CComediInput* pCIInput00 = new CComediInput();
	// Blocks are added to schedulers or other superblocks. In this case,
	// pSBAD1
	pSBAD1->AddComponent(pCIInput00);
	pCIInput00->SetIdentifier("input00");
	// Device file attached to the board to be used
	pCIInput00->SetDeviceFile(AD_COMEDI0);
	// Subdevice (functionality) of the board to be used
	pCIInput00->SetSubdevice(AD_ANALOG_INPUT);
	// Input to be read
	pCIInput00->SetChannel(0);
	// Range of the input voltage
	pCIInput00->SetRange(AD_RANGE_0_5);
	// Voltage reference for inputs
	pCIInput00->SetReference(AREF_GROUND);
	// The following two lines set the range of the output from 0 to 1. Since
	// the hardware range is set to AD_RANGE_0_5, 0V will be output as 0 and
	// +5V as 1.
	pCIInput00->SetOutputRange(1);
	pCIInput00->SetOutputOffset(0);

	// The following are Comedi Input Blocks for the rest of input
	// channels of the same A/D board. They are constructed as copies of
	// pCIInput00 and added to the same superblock, but then their own
	// identifier and channel number is set.
	CComediInput* pCIInput01 = new CComediInput(pCIInput00);
	pSBAD1->AddComponent(pCIInput01);
	pCIInput01->SetIdentifier("input01");
	pCIInput01->SetChannel(1);

	CComediInput * pCIInput02 = new CComediInput(pCIInput00);
	pSBAD1->AddComponent(pCIInput02);
	pCIInput02->SetIdentifier("input02");
	pCIInput02->SetChannel(2);

	CComediInput * pCIInput03 = new CComediInput(pCIInput00);
	pSBAD1->AddComponent(pCIInput03);
	pCIInput03->SetIdentifier("input03");
	pCIInput03->SetChannel(3);

	CComediInput * pCIInput04 = new CComediInput(pCIInput00);
	pSBAD1->AddComponent(pCIInput04);
	pCIInput04->SetIdentifier("input04");
	pCIInput04->SetChannel(4);

	CComediInput * pCIInput05 = new CComediInput(pCIInput00);
	pSBAD1->AddComponent(pCIInput05);
	pCIInput05->SetIdentifier("input05");
	pCIInput05->SetChannel(5);

	CComediInput * pCIInput06 = new CComediInput(pCIInput00);
	pSBAD1->AddComponent(pCIInput06);
	pCIInput06->SetIdentifier("input06");
	pCIInput06->SetChannel(6);

	CComediInput * pCIInput07 = new CComediInput(pCIInput00);
	pSBAD1->AddComponent(pCIInput07);
	pCIInput07->SetIdentifier("input07");
	pCIInput07->SetChannel(7);

	CComediInput * pCIInput08 = new CComediInput(pCIInput00);
	pSBAD1->AddComponent(pCIInput08);
	pCIInput08->SetIdentifier("input08");
	pCIInput08->SetChannel(8);

	CComediInput * pCIInput09 = new CComediInput(pCIInput00);
	pSBAD1->AddComponent(pCIInput09);
	pCIInput09->SetIdentifier("input09");
	pCIInput09->SetChannel(9);

	CComediInput * pCIInput10 = new CComediInput(pCIInput00);
	pSBAD1->AddComponent(pCIInput10);
	pCIInput10->SetIdentifier("input10");
	pCIInput10->SetChannel(10);

	CComediInput * pCIInput11 = new CComediInput(pCIInput00);
	pSBAD1->AddComponent(pCIInput11);
	pCIInput11->SetIdentifier("input11");
	pCIInput11->SetChannel(11);

	CComediInput * pCIInput12 = new CComediInput(pCIInput00);
	pSBAD1->AddComponent(pCIInput12);
	pCIInput12->SetIdentifier("input12");
	pCIInput12->SetChannel(12);

	CComediInput * pCIInput13 = new CComediInput(pCIInput00);
	pSBAD1->AddComponent(pCIInput13);
	pCIInput13->SetIdentifier("input13");
	pCIInput13->SetChannel(13);

	CComediInput * pCIInput14 = new CComediInput(pCIInput00);
	pSBAD1->AddComponent(pCIInput14);
	pCIInput14->SetIdentifier("input14");
	pCIInput14->SetChannel(14);

	CComediInput * pCIInput15 = new CComediInput(pCIInput00);
	pSBAD1->AddComponent(pCIInput15);
	pCIInput15->SetIdentifier("input15");
	pCIInput15->SetChannel(15);

	// This block joins the outputs of the previous 16 Comedi input blocks
	// in an array of type VARIABLES_TYPE and length AD_SE_CHANNELS.
	CFunction* pFInputsToArray = new CFunction();
	// It is part of the same pSBAD1 superblock
	pSBAD1->AddComponent(pFInputsToArray);
	pFInputsToArray->SetIdentifier("inputs_to_array");
	// Implementation of the operation. The name of the input and output
	// variables is set in their definition below
	pFInputsToArray->SetRunOperation(
		"inputs_array[0] = input00;\n"
		"inputs_array[1] = input01;\n"
		"inputs_array[2] = input02;\n"
		"inputs_array[3] = input03;\n"
		"inputs_array[4] = input04;\n"
		"inputs_array[5] = input05;\n"
		"inputs_array[6] = input06;\n"
		"inputs_array[7] = input07;\n"
		"inputs_array[8] = input08;\n"
		"inputs_array[9] = input09;\n"
		"inputs_array[10] = input10;\n"
		"inputs_array[11] = input11;\n"
		"inputs_array[12] = input12;\n"
		"inputs_array[13] = input13;\n"
		"inputs_array[14] = input14;\n"
		"inputs_array[15] = input15;\n"
		);
	// Inputs definitions
	pFInputsToArray->AddInput("input00", VARIABLES_TYPE);
	pFInputsToArray->AddInput("input01", VARIABLES_TYPE);
	pFInputsToArray->AddInput("input02", VARIABLES_TYPE);
	pFInputsToArray->AddInput("input03", VARIABLES_TYPE);
	pFInputsToArray->AddInput("input04", VARIABLES_TYPE);
	pFInputsToArray->AddInput("input05", VARIABLES_TYPE);
	pFInputsToArray->AddInput("input06", VARIABLES_TYPE);
	pFInputsToArray->AddInput("input07", VARIABLES_TYPE);
	pFInputsToArray->AddInput("input08", VARIABLES_TYPE);
	pFInputsToArray->AddInput("input09", VARIABLES_TYPE);
	pFInputsToArray->AddInput("input10", VARIABLES_TYPE);
	pFInputsToArray->AddInput("input11", VARIABLES_TYPE);
	pFInputsToArray->AddInput("input12", VARIABLES_TYPE);
	pFInputsToArray->AddInput("input13", VARIABLES_TYPE);
	pFInputsToArray->AddInput("input14", VARIABLES_TYPE);
	pFInputsToArray->AddInput("input15", VARIABLES_TYPE);
	// Every one of these inputs is connected to the output of one of the
	// previous Comedi input blocks
	pFInputsToArray->Input("input00")->Connect(pCIInput00->Output());
	pFInputsToArray->Input("input01")->Connect(pCIInput01->Output());
	pFInputsToArray->Input("input02")->Connect(pCIInput02->Output());
	pFInputsToArray->Input("input03")->Connect(pCIInput03->Output());
	pFInputsToArray->Input("input04")->Connect(pCIInput04->Output());
	pFInputsToArray->Input("input05")->Connect(pCIInput05->Output());
	pFInputsToArray->Input("input06")->Connect(pCIInput06->Output());
	pFInputsToArray->Input("input07")->Connect(pCIInput07->Output());
	pFInputsToArray->Input("input08")->Connect(pCIInput08->Output());
	pFInputsToArray->Input("input09")->Connect(pCIInput09->Output());
	pFInputsToArray->Input("input10")->Connect(pCIInput10->Output());
	pFInputsToArray->Input("input11")->Connect(pCIInput11->Output());
	pFInputsToArray->Input("input12")->Connect(pCIInput12->Output());
	pFInputsToArray->Input("input13")->Connect(pCIInput13->Output());
	pFInputsToArray->Input("input14")->Connect(pCIInput14->Output());
	pFInputsToArray->Input("input15")->Connect(pCIInput15->Output());
	// Output definition.
	pFInputsToArray->AddOutput("inputs_array", VARIABLES_TYPE, AD_SE_CHANNELS);
	// Superblock outputs should be connected to outputs of directly-contained
	// blocks or superblocks, to get their values from them.
	pSBAD1->Output("inputs_array")->Connect(
		pFInputsToArray->Output("inputs_array"));
	// The same connection can be made in the opposite way, like this:
	// pFInputsToArray->Output("inputs_array")->Connect(
	//   pSBAD1->Output("inputs_array"));
	// It's still the superblock output who gets the value from the subblock
	// output.

	// Logging fifo for the output of pSBAD1. A Logging fifo writes the value
	// of its containing variable to a "/dev/rtf*" file in every iteration of
	// the system. The file can be read by an user-space program. The values
	// are exactly in the same format as the variable, in this case arrays
	// of type VARIABLES_TYPE and length DAS_SE_CHANNELS.
	CLoggingFifo* pLFInputsArray = new CLoggingFifo();
	pLFInputsArray->SetIdentifier("logging_fifo");
	// Logging fifos are added to variables. Inputs and outputs are variables.
	// In this case, the output of pSBAD1.
	pSBAD1->Output("inputs_array")->AddComponent(pLFInputsArray);
	// The file will be "/dev/rtf1". "SetNumber(2)" would use "/dev/rtf2",
	// "SetNumber(3)" "/dev/rtf3", and so on.
	pLFInputsArray->SetNumber(1);
	// To reduce the SIGNIFICANT WORKLOAD OF REALTIME FIFOS, we can log in a
	// fraction of the iterations. The following means that the value of the
	// variables will be written in one every 1/LOGGING_SAMPLE_RATE iterations
	// of the scheduler.
	// NOT IMPLEMENTED YET
	// bNoError = pLFInputsArray->SetSamplingRate(LOGGING_SAMPLING_RATE);
	// Fifos work as buffers between realtime and user-space processes, so
	// they can contain several samples of the variable. The following sets
	// the maximum capacity to FIFOS_LENGTH samples.
	pLFInputsArray->SetLength(FIFOS_LENGTH);

#ifdef AD2

	// The following superblock is constructed as a copy of pSBAD1, so it does
	// exactly the same.
	CSuperBlock* pSBAD2 = new CSuperBlock(pSBAD1);
	pPSScheduler->AddComponent(pSBAD2);
	// The identifier must be different.
	pSBAD2->SetIdentifier("ad2");

	// However, we want the superblock to work with AD_COMEDI1 instead of
	// AD_COMEDI0. In order to get that, we have to set the device file in
	// everyone of the contained comedi input blocks. We don't have pointers
	// to those blocks, but we can obtain them by using the "Component" member
	// function with the identifier of the desired component as the first
	// argument and its class identifier as the second, as follows:
	((CComediInput *) pSBAD2->Component("input00", "CComediInput"))->
		SetDeviceFile(AD_COMEDI1);
	((CComediInput *) pSBAD2->Component("input01", "CComediInput"))->
		SetDeviceFile(AD_COMEDI1);
	((CComediInput *) pSBAD2->Component("input02", "CComediInput"))->
		SetDeviceFile(AD_COMEDI1);
	((CComediInput *) pSBAD2->Component("input03", "CComediInput"))->
		SetDeviceFile(AD_COMEDI1);
	((CComediInput *) pSBAD2->Component("input04", "CComediInput"))->
		SetDeviceFile(AD_COMEDI1);
	((CComediInput *) pSBAD2->Component("input05", "CComediInput"))->
		SetDeviceFile(AD_COMEDI1);
	((CComediInput *) pSBAD2->Component("input06", "CComediInput"))->
		SetDeviceFile(AD_COMEDI1);
	((CComediInput *) pSBAD2->Component("input07", "CComediInput"))->
		SetDeviceFile(AD_COMEDI1);
	((CComediInput *) pSBAD2->Component("input08", "CComediInput"))->
		SetDeviceFile(AD_COMEDI1);
	((CComediInput *) pSBAD2->Component("input09", "CComediInput"))->
		SetDeviceFile(AD_COMEDI1);
	((CComediInput *) pSBAD2->Component("input10", "CComediInput"))->
		SetDeviceFile(AD_COMEDI1);
	((CComediInput *) pSBAD2->Component("input11", "CComediInput"))->
		SetDeviceFile(AD_COMEDI1);
	((CComediInput *) pSBAD2->Component("input12", "CComediInput"))->
		SetDeviceFile(AD_COMEDI1);
	((CComediInput *) pSBAD2->Component("input13", "CComediInput"))->
		SetDeviceFile(AD_COMEDI1);
	((CComediInput *) pSBAD2->Component("input14", "CComediInput"))->
		SetDeviceFile(AD_COMEDI1);
	((CComediInput *) pSBAD2->Component("input15", "CComediInput"))->
		SetDeviceFile(AD_COMEDI1);

	// We also want the logging fifo to log to "/dev/rtf2" instead of
	// "/dev/rtf1". Notice inputs and outputs are components of their block,
	// although they are constructed and added in a single "AddInput" or
	// "AddOutput" call. Their class identifiers are "CInput" and "COutput".
	// So we need a double component search, one to get the output from the
	// superblock, and one to get fifo from the output:
	((CLoggingFifo *) ((COutput *) pSBAD2->Component(
		"inputs_array", "COutput"))->Component("logging_fifo", "CLoggingFifo"))->
		SetNumber(2);

#endif	// AD2

	// This is an example of a block that generates saw waves with maximum,
	// minimum, rising and falling slope determined by the A/D inputs.
	CFunction* pFFunction = new CFunction();
	// This block is added to pPSScheduler
	pPSScheduler->AddComponent(pFFunction);
	pFFunction->SetIdentifier("function");
	// This is the code that implements the function. It generates a saw
	// wave in every position of the output array "daoutputs".
	// For every position I from 0 to 3:
	// - ad1inputs[I*4] is the raising slope
	// - ad1inputs[I*4+1] is the maximum
	// - ad1inputs[I*4+2] is the falling slope
	// - ad1inputs[I*4+3] is the minimum
	// For every position I from 4 to 7:
	// - ad2inputs [(I-4)*4] is the raising slope
	// - ad2inputs [(I-4)*4+1] is the maximum
	// - ad2inputs [(I-4)*4+2] is the falling slope
	// - ad2inputs [(I-4)*4+3] is the minimum
	// Notice the code assumes AD_SE_CHANNELS >= 16 and DA_CHANNELS >=8
	pFFunction->SetBeforeRunOperation(
		// Array that holds the status of every wave
		"counter[0] = 0;\n"
		"counter[1] = 0;\n"
		"counter[2] = 0;\n"
		"counter[3] = 0;\n"
		"counter[4] = 0;\n"
		"counter[5] = 0;\n"
		"counter[6] = 0;\n"
		"counter[7] = 0;\n"
		// Boolean flags that determine if the slope si raising or falling.
		"increasing[0] = 0;\n"
		"increasing[1] = 0;\n"
		"increasing[2] = 0;\n"
		"increasing[3] = 0;\n"
		"increasing[4] = 0;\n"
		"increasing[5] = 0;\n"
		"increasing[6] = 0;\n"
		"increasing[7] = 0;\n"
		);
	pFFunction->SetRunOperation(
		// For the first four outputs
		"for (index = 0; index < 4; index++)\n"
		"{\n"
			// If we are in the rising slope
		"	if (increasing[index])\n"
		"	{\n"
				// ad1inputs[index*4] is the increment
		"		counter[index] = counter[index] + ad1inputs[index*4];\n"
				// If the maximum limit (ad1inputs[index*4+1]) is crossed
		"		if (counter[index] > ad1inputs[index*4+1])\n"
		"		{\n"
					// Move to the limit
		"			counter[index] = ad1inputs[index*4+1];\n"
					// Start the falling slope
		"			increasing[index] = 0;\n"
		"		}\n"
		"	}\n"
			// If we are in the falling slope
		"	else\n"
		"	{\n"
				// ad1inputs[index*4+2] is the decrement
		"		counter[index] = counter[index] - ad1inputs[index*4+2];\n"
				// If the minimum limit (ad1inputs[index*2+3]) is crossed
		"		if (counter[index] < ad1inputs[index*4+3])\n"
		"		{\n"
					// Set to the limit
		"			counter[index] = ad1inputs[index*4+3];\n"
					// Start the rising slope
		"			increasing[index] = 1;\n"
		"		}\n"
		"	}\n"
				// Set the output value
		"		daoutputs[index] = counter[index];\n"
		"}\n"

#ifdef AD2

		// For the following four outputs
		"for (index = 0; index < 4; index++)\n"
		"{\n"
		// If we are in the rising slope
		"	if (increasing[index+4])\n"
		"	{\n"
				// ad2inputs[index*4] is the increment
		"		counter[index+4] = counter[index+4] + ad2inputs[index*4];\n"
				// If the maximum limit (ad2inputs[index*4+1]) is crossed
		"		if (counter[index+4] > ad2inputs[index*4+1])\n"
		"		{\n"
					// Set to the limit
		"			counter[index+4] = ad2inputs[index*4+1];\n"
					// Start the falling slope
		"			increasing[index+4] = 0;\n"
		"		}\n"
		"	}\n"
			// If we are in the falling slope
		"	else\n"
		"	{\n"
				// ad2inputs[index*4+2] is the decrement
		"		counter[index+4] = counter[index+4] - ad2inputs[index*4+2];\n"
				// If the minimum limit (ad2inputs[index*4+3]) is crossed
		"		if (counter[index+4] < ad2inputs[index*4+3])\n"
		"		{\n"
					// Set to the limit
		"			counter[index+4] = ad2inputs[index*4+3];\n"
					// Start the rising slope
		"			increasing[index+4] = 1;\n"
		"		}\n"
		"	}\n"
			// Set output value
		"	daoutputs[index+4] = counter[index+4];\n"
		"}\n"

#endif

		);

	// Block local variables are be declared in the same way than
	// inputs and outpus
	pFFunction->AddBlockVariable("counter", VARIABLES_TYPE, DA_CHANNELS);
	pFFunction->AddBlockVariable("increasing", VARIABLES_TYPE, DA_CHANNELS);
	pFFunction->AddBlockVariable("index", "int");
	// Array of inputs connected to pSBAD1 output
	pFFunction->AddInput("ad1inputs", VARIABLES_TYPE, AD_SE_CHANNELS);
	pFFunction->Input("ad1inputs")->Connect(
		pSBAD1->Output("inputs_array"));

#ifdef AD2

	// Array of inputs connected to pSBAD1 output
	pFFunction->AddInput("ad2inputs", VARIABLES_TYPE, AD_SE_CHANNELS);
	pFFunction->Input("ad2inputs")->Connect(
		pSBAD2->Output("inputs_array"));

#endif

  // Array of outputs
	pFFunction->AddOutput("daoutputs", VARIABLES_TYPE, DA_CHANNELS);

#ifdef DDA

	// This superblock does the opposite than pSBAD1 and pSBAD2. It reads
	// an array of inputs and writes the values to the output channels of
	// DA_COMEDI3
	CSuperBlock* pSBDDA0xC = new CSuperBlock();
	pPSScheduler->AddComponent(pSBDDA0xC);
	pSBDDA0xC->SetIdentifier("dda0xC");
	// Input array, connected to pFFunction's output.
	pSBDDA0xC->AddInput("outputs_array", VARIABLES_TYPE, DA_CHANNELS);
	pSBDDA0xC->Input("outputs_array")->Connect(
		pFFunction->Output("daoutputs"));

	// This block does the opposite than pFAD1ToArray above, splitting
	// an input array into several sipmle outputs.
	CFunction* pFArrayToOutputs = new CFunction();
	pSBDDA0xC->AddComponent(pFArrayToOutputs);
	pFArrayToOutputs->SetIdentifier("array_to_outputs");
	pFArrayToOutputs->SetRunOperation(
		"output0 = outputs_array[0];\n"
		"output1 = outputs_array[1];\n"
		"output2 = outputs_array[2];\n"
		"output3 = outputs_array[3];\n"
		"output4 = outputs_array[4];\n"
		"output5 = outputs_array[5];\n"
		"output6 = outputs_array[6];\n"
		"output7 = outputs_array[7];\n"
		);
	// Array of inputs.
	pFArrayToOutputs->AddInput("outputs_array", VARIABLES_TYPE, DA_CHANNELS);
	// Block inputs can be connected to inputs of its superblock, to get
	// their values from them. In this case:
	pFArrayToOutputs->Input("outputs_array")->Connect(
		pSBDDA0xC->Input("outputs_array"));
	// The same connection can be made in the opposite way, like this:
	// pSBDDA0xC->Input("outputs_array")->Connect(
	//   pFArrayToOutputs->Input("outputs_array"));
	// It's still the block input who gets the value from the superblock
	// input.

	// Outputs for every position of the array
	pFArrayToOutputs->AddOutput("output0", VARIABLES_TYPE);
	pFArrayToOutputs->AddOutput("output1", VARIABLES_TYPE);
	pFArrayToOutputs->AddOutput("output2", VARIABLES_TYPE);
	pFArrayToOutputs->AddOutput("output3", VARIABLES_TYPE);
	pFArrayToOutputs->AddOutput("output4", VARIABLES_TYPE);
	pFArrayToOutputs->AddOutput("output5", VARIABLES_TYPE);
	pFArrayToOutputs->AddOutput("output6", VARIABLES_TYPE);
	pFArrayToOutputs->AddOutput("output7", VARIABLES_TYPE);

	// The following are the declarations of Comedi output blocks, one for
	// every output channel of the D/A board. They are like Comedi input
	// blocks, only that they connect to other blocks with an input
	// instead of an output (obviously, to get the value to be written to
	// the hardware output)
	CComediOutput *pCOOutput0 = new CComediOutput();
	pSBDDA0xC->AddComponent(pCOOutput0);
	pCOOutput0->SetIdentifier("output0");
	pCOOutput0->SetDeviceFile(DA_COMEDI3);
	pCOOutput0->SetSubdevice(DA_ANALOG_OUTPUT);
	pCOOutput0->SetChannel(0);
	pCOOutput0->SetRange(DA_RANGE_0_5);
	pCOOutput0->SetReference(AREF_GROUND);
	// The following two lines set the range of the input from 0 to 1. Since
	// the hardware range is set as DA_RANGE_0_5, 0 will output 0V and
	// 1 +5V.
	pCOOutput0->SetInputRange(1);
	pCOOutput0->SetInputOffset(0);
	pCOOutput0->Input()->Connect(pFArrayToOutputs->Output("output0"));

	// The rest of Comedi output blocks are copy-constructed as it was done
	// with the Comedi input blocks.
	CComediOutput *pCOOutput1 = new CComediOutput(pCOOutput0);
	pSBDDA0xC->AddComponent(pCOOutput1);
	pCOOutput1->SetIdentifier("output1");
	pCOOutput1->SetChannel(1);
	pCOOutput1->Input()->Connect(pFArrayToOutputs->Output("output1"));

	CComediOutput *pCOOutput2 = new CComediOutput(pCOOutput0);
	pSBDDA0xC->AddComponent(pCOOutput2);
	pCOOutput2->SetIdentifier("output2");
	pCOOutput2->SetChannel(2);
	pCOOutput2->Input()->Connect(pFArrayToOutputs->Output("output2"));

	CComediOutput *pCOOutput3 = new CComediOutput(pCOOutput0);
	pSBDDA0xC->AddComponent(pCOOutput3);
	pCOOutput3->SetIdentifier("output3");
	pCOOutput3->SetChannel(3);
	pCOOutput3->Input()->Connect(pFArrayToOutputs->Output("output3"));

	CComediOutput *pCOOutput4 = new CComediOutput(pCOOutput0);
	pSBDDA0xC->AddComponent(pCOOutput4);
	pCOOutput4->SetIdentifier("output4");
	pCOOutput4->SetChannel(4);
	pCOOutput4->Input()->Connect(pFArrayToOutputs->Output("output4"));

	CComediOutput *pCOOutput5 = new CComediOutput(pCOOutput0);
	pSBDDA0xC->AddComponent(pCOOutput5);
	pCOOutput5->SetIdentifier("output5");
	pCOOutput5->SetChannel(5);
	pCOOutput5->Input()->Connect(pFArrayToOutputs->Output("output5"));

	CComediOutput *pCOOutput6 = new CComediOutput(pCOOutput0);
	pSBDDA0xC->AddComponent(pCOOutput6);
	pCOOutput6->SetIdentifier("output6");
	pCOOutput6->SetChannel(6);
	pCOOutput6->Input()->Connect(pFArrayToOutputs->Output("output6"));

	CComediOutput *pCOOutput7 = new CComediOutput(pCOOutput0 );
	pSBDDA0xC->AddComponent(pCOOutput7);
	pCOOutput7->SetIdentifier("output7");
	pCOOutput7->SetChannel(7);
	pCOOutput7->Input()->Connect(pFArrayToOutputs->Output("output7"));

#endif	// DDA

	// Once the definition of the dynamic system is complete, ArrangeAll must
	// be called.
	bool bNoError = DynamicSystem.ArrangeAll();

	// If any error has been detected in the definition...
	if (!bNoError)
	{
		CMessages Messages;
		DynamicSystem.AllReportMessages(Messages);
		CMessages::m_TMessagesVector::const_iterator MessagesIterator;

		// The information about the detected errors is printed.
		for (MessagesIterator = Messages.Vector().begin();
			MessagesIterator != Messages.Vector().end();
			MessagesIterator++)
		{
			std::cout << "Error Number: " << (*MessagesIterator).Number() <<
				std::endl;
			std::cout << "Class with the error: " <<
				(*MessagesIterator).Location()->Class() << std::endl;
			std::cout << "Identifier of the class with the error: " <<
				(*MessagesIterator).Location()->Identifier() << std::endl;
			if ((*MessagesIterator).Location()->Container() != NULL)
				std::cout << "Identifier of the container of the class with the "
					"error: " <<
					(*MessagesIterator).Location()->Container()->Identifier() <<
						std::endl;
		}
	}
	// If there are no errors
	else
	{
		// The realtime implementation of the system is writen to
		// ".rtcode/rtmodule.c"
		std::ofstream FileStream;
		FileStream.open("./rtcode/rtmodule.c");
		// The generation templates are in "./dslib-2/templates/rtai_module/"
		DynamicSystem.GenerateCode("./dslib-2/templates/rtai_module/", FileStream);
		FileStream.close();
	}
	// The following deletes all the objects contained by DynamicSystem,
	// directly or indirectly, so all the allocated memory is freed.
	DynamicSystem.DeleteObjects();

	return EXIT_SUCCESS;
}
